#!/usr/bin/perl

# Author : Damien KROTKINE (dams@tuxfamily.org)
#
# Contributors : 
# Charles LONGEAU (chl@tuxfamily.org)
#
# Copyright (C) 2002 Damien KROTKINE (dams@tuxfamily.org)
# Copyright (C) 2002 Charles LONGEAU (chl@tuxfamily.org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package Libconf;

=head1 NAME

  Libconf - An abstraction layer to config files

=head1 SYNOPSIS

=head2 powerfull perl structure

  If you plan to access the config files using perl only, use Libconf::Glueconf :

  $samba = new Libconf::Glueconf::Samba('/etc/samba/smb.conf');
  $samba->readConf();
  use Data::Dumper;
  print Dumper($samba) . "\n";
  $samba->{global}{workgroup} = 'WORKGROUP';
  $samba->{global}{'netbios name'} = 'some other name';
  $samba->{share} = {
      comment => 'Share directory',
      path => '/mnt/share',
      public => 1,
      writable => 0,
  };
  delete $samba->{old_share};
  $samba->writeConf();

  See Libconf::Glueconf for list of modules, and details

=head2 XML access

  If you want to use the XML representation of the config file :

  use Libconf;
  my $samba = new Libconf('/etc/samba/smb.conf', 'samba', '');

  # we want the xml representation of the Libconf object
  my $samba_xmlstring = $samba->toXMLString();

  # we'll edit it using XML::Twig
  require XML::Twig;
  my $samba_twig= XML::Twig->new();
  $samba_twig->parse($samba_xmlString);
  my $root = $samba_twig->root;
  my $element = $root->first_child('CONFIG-FILE');

  # we want to change the name of the section 'public' to 'private'

  my $section = $element->children('SECTION[@name="public"]');
  $section->{att}->{name} = 'private';

  $samba = newFromXML($samba_twig->toString());
  $samba->writeConf('/etc/samba/smb.conf');

=head2 Libconf internal

  You could use this if there is no Glueconf module, but in this case, you better mail us so that we implement the glueconf module

  $struct = new Libconf('/etc/samba/smb.conf', 'samba', '');
  $struct->editAtom(-1, {
                          key => 'writable',
                          value => 'no',
                          sections => [{name => 'homes'}] 
                        }
                   );
  $struct->appendAtom( {
                         key => 'key_name',
                         value => 'value',
                         type => 'KEY_VALUE'
                       } );
  $struct->writeConf('/etc/samba/smb.conf');

=head1 DESCRIPTION

Libconf is a powerfull set of modules used to access config files. Its advantages are

=over

=item *

libconf doesn't loose any config information, doesn't overwrite informations

=item *

a powerfull internal parser that respects comments of any kind (single/multiple characters, with begin and end, and with nested comments)

=item *

Glueconf, a special module to access the config information of a single config file very easily in perl

=item *

Systemconf, a special module to access the config information of a configuration part of the system, you don't need to know which config files are affected. The code written using systemconf is meant to be distribution independant.

=item *

XML interaction, so that you can use anything that can handle XML to edit your config files. That could be some xsl, C LibXML or anything else.

=back

=cut


use strict;
use vars qw(@ISA %EXPORT_TAGS @EXPORT_OK $VERSION $INDENT_SPACES $TEMPLATES_DIRECTORY %templates );

@ISA = qw(Exporter);
%EXPORT_TAGS = (
    functions => [ qw() ],
    settings => [ qw() ],
);
$EXPORT_TAGS{all} = [ map { @$_ } values %EXPORT_TAGS ];
@EXPORT_OK = map { @$_ } values %EXPORT_TAGS;

$VERSION='0.27';

$INDENT_SPACES = '    ';

=head2 new()


=cut

sub new {
    my ($pkg, $filename, $template_name, $validation_name, $template_filename, $safe_write) = @_;
#    -e $filename or die ("$filename not found");
    defined $filename && !-e $filename and system("touch", $filename);
    if ($template_filename) {
        require $template_filename;
    } else {
        eval "require Libconf::templates::$template_name" . '_template';
        $@ and die "could not require Libconf::templates::$template_name" . '_template : ' . $@;
    }
    my %out;
    $out{filename} = $filename;
    $out{template_name} = $template_name;
    $validation_name and $out{validation_name} = $validation_name;
    $out{atoms} = [];
    defined $filename && $safe_write and $out{timestamp} = getTimestamp($filename);
    readConf(\%out);
    bless \%out, $pkg;
    \%out;
}

sub getTimestamp {
    my ($filename) = @_;
    my @stat;
    @stat = stat($filename);
    return $stat[9];
}

=head2 readConf()

=cut

sub readConf {
    my ($out) = @_;
    my $filename = $out->{filename};
    my @file;
    if (-e $filename) {
        @file = cat_($filename);
    } else {
        @file = ();
    }
    parse($out, \@file);
}

sub clear {
    my ($out) = @_;
    $out->{atoms} = [];
}

sub size {
    my ($out) = @_;
    scalar(@{$out->{atoms}});
}

# COMMENTS TO BE REVIEWED
# $out : the structure
# $index : the atom index. If -1, then the template will try to find the matching atom
# $ref_args : a reference on a hash containing the datas
# returns : 
#   0 if ok,
#   -1 if $index bigger than @atoms,
#   -2 if the template didn't succeed finding the mathing atom ($index was -1),
#   -3 if there is an error setting the values in the atom

=head2 editAtom()


=cut

sub editAtom {
    my ($out, $index, $ref_args) = @_;
    my %args=%$ref_args;
    eval($templates{$out->{template_name}}{edit_atom});
    $@ and die $@;
#    validate($out, @{$out->{atoms}}[$index]);
}

=head2 getAtom()


=cut

sub getAtom {
    my ($out, $index) = @_;
    $out->{atoms}[$index];
}


=head2 findAtomPos()


=cut

sub findAtomPos {
    my ($out, $ref_args, $first_atom, $last_atom) = @_;
    my %args=%$ref_args;
    eval($templates{$out->{template_name}}{find_atom_pos});
}


=head2 appendAtom()


=cut

sub appendAtom {
    my ($out, $ref_args) = @_;
    my %args = %$ref_args;
    my $index = @{$out->{atoms}};
    push(@{$out->{atoms}}, {});
    eval($templates{$out->{template_name}}{edit_atom});
#    validate($out, @{$out->{atoms}}[$index]);
}

=head2 insertAtom()


=cut

sub insertAtom {
    my ($out, $index, $ref_args) = @_;
    $index >= @{$out->{atoms}} and return appendAtom($out, $ref_args);
    my %args = %$ref_args;
    splice(@{$out->{atoms}}, $index, 0, {});
    $args{sections} ||= @{$out->{atoms}}[$index-1]->{sections};
    $args{sections} ||= [];
    eval($templates{$out->{template_name}}{edit_atom});
#    validate($out, @{$out->{atoms}}[$index]);
}

=head2 deleteAtom()


=cut

sub deleteAtom {
    my ($out, $index) = @_;  
    splice(@{$out->{atoms}}, $index, 1);
#    validate($out, @{$out->{atoms}}[$index]);
}

sub compare_depth {
    my ($ref1, $ref2) = @_;
    my @sections1 = @$ref1;
    my @sections2 = @$ref2;
    @sections1 == @sections2 or return 0;
    foreach my $i (0..@sections1-1) {
        @sections2[$i]->{name} eq @sections1[$i]->{name} or return 0;
    }
    1;
}

sub deleteSection {
    my ($out, $index, $no_end_section) = @_;
    my $atom = $out->getAtom($index);
    #$atom->sections;
    my @end_positions = $no_end_section ?
      $out->findAtomPos( { type => 'SECTION' }) :
        $out->findAtomPos( { type => 'ENDSECTION', sections => $atom->{sections} });
    my $end;
    foreach (@end_positions) {
        $_ > $index and $end = $_, last;
    }
    !defined $end && $no_end_section and $end = @{$out->{atoms}}-1;
    defined $end or die "no ENDSECTION for SECTION n $index";
    splice(@{$out->{atoms}}, $index, $end-$index+1);
}

=head2 clearSection()


=cut

sub clearSection {
    my ($out, $index, $no_end_section) = @_;
    my $atom = $out->getAtom($index);
    #$atom->sections;
    my @end_positions = $no_end_section ?
      $out->findAtomPos( { type => 'SECTION' }) :
        $out->findAtomPos( { type => 'ENDSECTION', sections => $atom->{sections} });
    my $end;
    foreach (@end_positions) {
        $_ > $index and $end = $_, last;
    }
    !defined $end && $no_end_section and $end = @{$out->{atoms}}-1;
    defined $end or die "no ENDSECTION for SECTION n $index";
    my @ret = splice(@{$out->{atoms}}, $index+1, $end-$index);
    @ret;
}

=head2 writeConf()

=cut

sub writeConf {
    my ($out, $filename) = @_;
    $filename ||= $out->{filename};
    my $template_name = $out->{template_name};
    my @out_atoms = @{$out->{atoms}};
    if (defined $out->{timestamp} && (getTimestamp($filename) != $out->{timestamp})) {
        print "Warning $filename has changed !\n";
    }
    local *F;
    open F, "> $filename" or die "cannot create config file $filename (do you have proper permissions?)";
    foreach my $atom (@out_atoms) {
#	print F "\n" x $$atom{line_feed};
#	$$atom{comments} and print F eval($templates{$template_name}{comment_output}) foreach split("\n", $$atom{comments});
        my $output_text = '';
        my $output_indentation = $out->{current_indentation};
        if (exists $templates{$template_name}->{output}{$atom->{type}}) {
            eval($templates{$template_name}->{output}{$atom->{type}});
        } else {
            if (defined($atom->{type})) {
                print F '## WARNING - No valid $template->{output}{' . $atom->{type} . "}\n";
                die '## WARNING - No valid $template->{output}{' . $atom->{type} . "}\n";
            }
        }
        print F handle_output($output_text, $output_indentation, $atom->{comments});
    }
}

sub parse {
    my ($out, $file) = @_;
    my @file = @$file;
    my $item = 0;
    my $template_name = $out->{template_name};
    my @out_atoms;
    my $matched;
    my $keep_atom;
    my $switch_to_next_atom = 0;
    my $switch_to_prev_atom = 0;

    $out->{current_sections} ||= [];
    my $line_idx = 1;
    foreach my $in (@file) {
	$out_atoms[$item] ||= {};
        $out_atoms[$item]->{comments} ||= [];
	$in =~ /^\s*\n$/ and push(@{$out_atoms[$item]->{comments}}, { comment => '', inline => 0 }), next;
	chomp $in;
	#see if we are in a line previously commented
	if ($out_atoms[$item]->{current_comment}) {
	    my $c = $out_atoms[$item]->{current_comment};
	    if ($in =~ s/^(.*\Q$c\E)//) {
                push(@{$out_atoms[$item]->{comments}}, { comment => $1, inline => $in =~ /^\s*$/ ? 0 : -1 });
		undef $out_atoms[$item]->{current_comment};
	    } else {
                push(@{$out_atoms[$item]->{comments}}, { comment => $in, inline => 0 });
		$in = '';
	    }
	} else {
	    # we remove comments
	    # the comment structure : [ open_char, close_char, nested ]
	    my ($open, $close, $nested);
	    foreach (@{$templates{$template_name}{comments}}) {
		($open, $close, $nested) = @$_;
		if ($nested) {
		    my ($level, $start) = (0, 0);  #we don't use $out_atoms[$item]->{current_level} directly to avoid overriding it.
		    my $end = 0; #length($in);
		    for (my $i=0; $i < length($in); $i++) {
			substr($in, $i, length($open)) eq $open and $level++, $start ||= $i;
			if ($level && substr($in, $i, length($close)) eq $close) {
			    $level--;
			    $level or $end = $i-$start;
			}
		    }
		    $end and push(@{$out_atoms[$item]->{comments}}, { comment => substr($in, $start, $end + length($close)),
                                                                    inline => $start > 0 ? 1 : 0 });
		    substr($in, $start, $end + length($close)) = '';
		    $level and $out_atoms[$item]->{current_comment} = $close, $out_atoms[$item]->{current_level} = $level;
		} elsif (defined $open && defined $close) {
		    $in =~ s/(\Q$open\E.*\Q$close\E)/push(@{$out_atoms[$item]->{comments}}, { comment => $1, inline => length($') ? -1 : (length($`) ? 1 : 0) }); undef/eg;
		    #$in =~ s/(\Q$open\E.*\Q$close\E)/$out_atoms[$item]->{comments} .= "$1\n"; undef/eg;
		    $in =~ s/(\Q$open\E.*)$// and push(@{$out_atoms[$item]->{comments}}, { comment => $1,
                                                                                        inline => $in =~ /^\s*$/ ? 0 : 1 }
                                                      ), $out_atoms[$item]->{current_comment} = $close, last;
		} elsif (defined $open) {
                    if ($in =~ s/(\Q$open\E.*)$//) {
                        my $tmp = $1;
                        push(@{$out_atoms[$item]->{comments}}, { comment => $tmp,
                                                                 inline => $in =~ /^\s*$/ ? 0 : 1 });
                    }
		}
	    }
        }
	### End Comment handling ###
        $matched = 0;
        $keep_atom = 0;
        $switch_to_next_atom = 0;
        $switch_to_prev_atom = 0;
        my $atom = $out_atoms[$item];
        foreach (@{$templates{$template_name}{rules}}) {
	    eval ($_);
	    $@ and die qq(error while applying the rule '$_' to '$in' line n $line_idx : $@);
            # the last rule asked to go to next atom
            if ($switch_to_next_atom) {
                $item++;
                $out_atoms[$item] ||= {};
                $out_atoms[$item]->{comments} ||= [];
                $atom = $out_atoms[$item];
                $switch_to_next_atom;
            }
            if ($switch_to_prev_atom) {
                $item--;
                $atom = $out_atoms[$item];
                $switch_to_prev_atom = 0;
            }
#            $next_item = $matched && !$keep_atom;
	    $matched and last;
	}
        if ($in !~ /^\s*$/) {
            push @{$out_atoms[$item]->{invalid}}, $in;
            print "\nLibconf.pm warning: can't parse line n" . $line_idx . " :\n$in\n";
        }
        exists $atom->{sections} or $atom->{sections} = [];
        $matched && !$keep_atom and $item++;
    } continue {
        $line_idx++;
    }
    $out->{atoms} = [@out_atoms];
}

sub handle_output {
    my ($text, $indentation, $comments) = @_;
    my $ret = '';
    my $flag = 1;
    my $same_line = 0;
    $indentation = $INDENT_SPACES x $indentation;
    if (defined $comments) {
        for (my $i; $i < @$comments; $i++) {
            my $comment = $comments->[$i];
            if ($comment->{inline} == -1) {
                $ret .= $indentation . $comment->{comment};
                $same_line = 1;
                if ($i == @$comments-1) {
                    $ret .= " $text" . "\n";
                    $flag = 0;
                }
            } elsif ($comment->{inline} == 0) {
                $ret .= $indentation . $comment->{comment} . "\n";
            } elsif ($comment->{inline} == 1) {
                if ($same_line) {
                    $same_line = 0;
                } else {
                    $ret .= $indentation;
                }
                if ($i == @$comments-1) {
                    $ret .= "$text " . $comment->{comment} . "\n";
                    $flag = 0;
                } else {
                    $ret .= $comment->{comment} . "\n";
                }
            }
        }
    }
    $flag && length($text) and $ret .= $indentation . $text . "\n";
    $ret;
}

#validation stuff

sub setUniq {
    my ($obj) = @_;
    my @to_delete;
    my $i = 0;
    foreach my $atom (@{$obj->{atoms}}) {
        my %atom2 = %$atom;
        delete @atom2{qw(comments value list)};
        my @pos = findAtomPos($obj, \%atom2);
        if (@pos > 1 && $pos[-1] != $i) {
            push @to_delete, $i;
        }
        $i++;
    }
    print "@to_delete\n";
    deleteAtom($obj, $_) foreach reverse @to_delete;
}

sub debug {
    $ENV{DEBUG} or return;
    my $i = 0;
    while (caller($i)) { $i++ }
    my ($_package, $_filename, $_line, $subroutine, $_hasargs, $_wantarray, $_evaltext, $_is_require, $_hints, $_bitmask) = caller(2);
    print '   ' x ($i-3);
    print "\033[35m";
    print $subroutine, "\t\033[36m", @_;
    print "\033[0m\n";
}

# sub validate {
#     my ($out, $atom) = @_;
#     $out->{validation_name} and eval($validation{$out->{validation_name}}->{rules});
# }

sub cat_ { local *F; open F, $_[0] or return; my @l = <F>; wantarray() ? @l : join '', @l }
sub output { my $f = shift; local *F; open F, ">$f" or die "output in file $f failed: $!\n"; print F foreach @_; 1 }
sub member { my $e = shift; foreach (@_) { $e eq $_ and return 1 } 0 }
sub map_index(&@) {
    my $f = shift;
    my @v; local $::i = 0;
    map { @v = &$f($::i); $::i++; @v } @_;
}
sub if_($@) {
    my $b = shift;
    $b or return ();
    wantarray() || @_ <= 1 or die("if_ called in scalar context with more than one argument " . join(":", caller()));
    wantarray() ? @_ : $_[0];
}
sub fold_left(&@) {
    my ($f, $initial, @l) = @_;
    local ($::a, $::b);
    $::a = $initial;
    foreach (@l) { $::b = $_; $::a = &$f() }
    $::a
}

sub toXML {
    my ($self) = @_;
    require XML::LibXML;
    # new XML document
    my $doc = XML::LibXML::Document->new();
    $doc->setEncoding('UTF8');

    # element : config-file
    $doc->setDocumentElement(my $config_file = $doc->createElement('CONFIG-FILE'));
    $config_file->setAttribute(file_name => $self->{filename});
    $config_file->setAttribute(template_name => $self->{template_name});

    my @current_nodes = ($config_file);
    my $current_sections = [];


    foreach my $atom (@{$self->{atoms}}) {
        my $current_node = $current_nodes[-1];
        my ($child, $child2);
        defined $current_node or die "problem in xml generation or structure malformed";
        if      ($atom->{type} eq 'SECTION') {
            if (!compare_depth($atom->{sections}, $current_sections)) {
                pop @current_nodes;
                $current_node = $current_nodes[-1];
            };
            $current_node->appendChild($child = $doc->createElement('SECTION'));
            $child->setAttribute(name => $atom->{section_name});
            push @current_nodes, $child;
        } elsif ($atom->{type} eq 'KEY_VALUE') {
            $current_node->appendChild($child = $doc->createElement('KEY_VALUE'));
            $child->setAttribute(key => $atom->{key});
            $child->appendText($atom->{value});
        } elsif ($atom->{type} eq 'KEY_VALUES') {
            $current_node->appendChild($child = $doc->createElement('KEY_VALUES'));
            $child->setAttribute(key => $atom->{key});
            while (my ($key, $value) = each %{$atom->{values}}) {
                $child->appendChild($child2 = $doc->createElement('KEY_VALUE'));
                $child2->setAttribute(key => $key);
                $child2->appendText($value);
            }
        } elsif ($atom->{type} eq 'ENDSECTION') {
            pop @current_nodes;
        } elsif ($atom->{type} eq 'KEY_LIST') {
            $current_node->appendChild($child = $doc->createElement('KEY_LIST'));
            $child->setAttribute(key => $atom->{key});
            foreach my $value (@{$atom->{list}}) {
                $child->appendChild($child2 = $doc->createElement('VALUE'));
                $child2->appendText($value);
            }
        }
        if (exists $atom->{comments} and @{$atom->{comments}}) {
            defined $child or $current_node->appendChild($child = $doc->createElement('VOID'));
            $child->appendChild($child2 = $doc->createElement('COMMENTS'));
            foreach my $comment (@{$atom->{comments}}) {
                $child2->appendChild(my $child3 = $doc->createElement('COMMENT'));
                $child3->appendText($comment->{comment});
                $comment->{inline} and $child3->setAttribute(inline => $comment->{inline} == 1 ? 'RIGHT' : 'LEFT' );
            }
        }
        exists $atom->{sections} and $current_sections = $atom->{sections};
    }
    $doc;
}

sub toXMLString {
    my ($self, $format) = @_;
    @{toXML($self)->getElementsByTagName('CONFIG-FILE')}[0]->toString($format);
}

sub toXMLFile {
    my ($self, $file, $format) = @_;
    toXML($self)->toFile($file, $format);
}


sub convertComments {
    my ($elt, $array) = @_;
    foreach ($elt->children('COMMENTS')) {
        foreach ($_->children('COMMENT')) {
            push @$array, {comment => $_->text};
        }
    }
}

# sub newFromXML {
#     my ($pkg, $xmlString) = @_;
#     require XML::Twig;
#     my $twig= XML::Twig->new();
#     $twig->parse($xmlString);
#     my $elt = $twig->root;
#     my %self;
#     $self{template_name} = $elt->{att}->{template_name};
#     $self{filename} = $elt->{att}->{file_name};
#     foreach ($elt->children('SECTION')) {
#         my %section;
#         $section{type} = 'SECTION';
#         $section{section_name} = $_->{att}->{name};
# 	$section{sections} = [];
# 	$self{current_sections} = [{name => $_->{att}->{name}}];
#         push @{$self{atoms}}, \%section;
#         convertComments($_, \@{$section{comments}});
#         foreach ($_->children('KEY_LIST')) {
#             my %key_list;
#             convertComments($_, \@{$key_list{comments}});
#             $key_list{type} = 'KEY_LIST';
#             $key_list{key} = $_->{att}->{key};
# 	    $key_list{sections} =  [ $self{current_sections}[0] ];
#             foreach ($_->children('VALUE')) {
#                 push @{$key_list{list}}, $_->text;
#             }
#             push @{$self{atoms}}, \%key_list;
#         }
# 	my %endsection;
# 	$endsection{type} = 'ENDSECTION';
#         push @{$self{atoms}}, \%endsection;
#     }
#     bless \%self, $pkg;
# }

sub newFromXML {
    my ($pkg, $xmlString) = @_;
    require XML::Twig;
    my $twig= XML::Twig->new();
    $twig->parse($xmlString);
    my $elt = $twig->root;
    my %self;
    $self{template_name} = $elt->{att}->{template_name};
    $self{filename} = $elt->{att}->{file_name};
    foreach ($elt->children('SECTION')) {
        my %section;
        $section{type} = 'SECTION';
        $section{section_name} = $_->{att}->{name};
	$section{sections} = [];
	$self{current_sections} = [{name => $_->{att}->{name}}];
        push @{$self{atoms}}, \%section;
        convertComments($_, \@{$section{comments}});
        foreach ($_->children('KEY_LIST')) {
            my %key_list;
            convertComments($_, \@{$key_list{comments}});
            $key_list{type} = 'KEY_LIST';
            $key_list{key} = $_->{att}->{key};
	    $key_list{sections} =  [ $self{current_sections}[0] ];
            foreach ($_->children('VALUE')) {
                push @{$key_list{list}}, $_->text;
            }
            push @{$self{atoms}}, \%key_list;
        }
	my %endsection;
	$endsection{type} = 'ENDSECTION';
        push @{$self{atoms}}, \%endsection;
    }
    bless \%self, $pkg;
}

sub recurseFromXml {
#    my () = @_;
}

=head1 SEE ALSO

Libconf

=cut

1;
